//Data structures for lls.cpp

// AVL trees implemented with a basis in 
// http://kukuruku.co/hub/cpp/avl-trees   (By Nikolai Ershov at the Russian Academy of Sciences in 2008.)

////////////////////////////////////////////////////////////////////////////////////////////////
//              The AVL tree for all the fibers within a perimeter
////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char getHeight(Fiber* f,Perimeter* p)
{   
    if(!f) return 0;
    for (int i = 0; i < 4; ++i)
    {
        if(f->myPerimeter[i]==p){
            return f->height[i];
        }
    }
    return 0;
}

Fiber* Fiber::getRight(Perimeter *p){
	for (int i = 0; i < 4; ++i){
    	if(myPerimeter[i]==p){
    		return right[i];

    	}
    }
    return NULL;
}

Fiber* Fiber::getLeft(Perimeter *p){
	for (int i = 0; i < 4; ++i){
    	if(myPerimeter[i]==p){
    		return left[i];

    	}
    }
    return NULL;
}

void Fiber::setRight(Fiber *f, Perimeter *p){
	for (int i = 0; i < 4; ++i){
    	if(myPerimeter[i]==p){
    		right[i]=f;
    		return;
    	}
    }
    return;
}

void Fiber::setLeft(Fiber *f, Perimeter *p){
	for (int i = 0; i < 4; ++i){
    	if(myPerimeter[i]==p){
    		left[i]=f;
    		return;
    	}
    }
    return;
}

int bfactor(Fiber* f,Perimeter *p)
{
    return getHeight(f->getRight(p),p)-getHeight(f->getLeft(p),p);
}

void fixHeight(Fiber* f,Perimeter *p)
{	
	for (int i = 0; i < 4; ++i)
    {
    	if(f->myPerimeter[i]==p){
    		unsigned char hl = getHeight(f->left[i],p);
    		unsigned char hr = getHeight(f->right[i],p);
    		f->height[i] = (hl>hr?hl:hr)+1;
    		break;
    	}
    }
}

Fiber* rotateright(Fiber* f,Perimeter *p)
{
    Fiber* q = f->getLeft(p);
    f->setLeft(q->getRight(p),p);
    q->setRight(f,p);
    fixHeight(f,p);
    fixHeight(q,p);
    return q;
}


Fiber* rotateleft(Fiber* q,Perimeter *p)
{
    Fiber* f = q->getRight(p);
    q->setRight(f->getLeft(p),p);
    f->setLeft(q,p);
    fixHeight(q,p);
    fixHeight(f,p);
    return f;
}

Fiber* balance(Fiber* f,Perimeter *p) // balancing the f Fiber
{
    fixHeight(f,p);
    if( bfactor(f,p)>=2 )
    {
        if( bfactor(f->getRight(p),p) < 0 )
            f->setRight(rotateright(f->getRight(p),p),p);
        return rotateleft(f,p);
    }
    if( bfactor(f,p)<=-2 )
    {
        if( bfactor(f->getLeft(p),p) > 0  )
            f->setLeft(rotateleft(f->getLeft(p),p),p);
        return rotateright(f,p);
    }
    return f; 
}
 
Fiber* insert(Fiber* f, Fiber* n, Perimeter *p) // insert fiber n in a tree with root f;
{
    if( !f ) return n;
    if( n->t < f->t )
        f->setLeft( insert(f->getLeft(p),n,p) ,p);
    else
    	f->setRight( insert(f->getRight(p),n,p),p);
    return balance(f,p);
}

Fiber* findmax(Fiber *f ,Perimeter *p){     //Fiber f is the root node of perimeter p 
	return f->getRight(p) ? findmax(f->getRight(p),p) : f;
}

Fiber* removemin(Fiber* f, Perimeter *p) // deleting a Fiber with minimal key from a f tree
{
    if( f->getLeft(p) ==NULL )
        return f->getRight(p);
    f->setLeft( removemin(f->getLeft(p),p) ,p);

    return balance(f,p);
}

Fiber* findmin(Fiber *f ,Perimeter *p){
	return f->getLeft(p) ? findmin(f->getLeft(p),p) : f;


}
Fiber* remove(Fiber* f, Fiber *o, Perimeter* p) // deleting old Fiber o from tree with root fiber f, within perimeter p
{	
    if( !f ) return 0;
    if( o->t < f->t )
        f->setLeft( remove(f->getLeft(p),o,p) ,p); 
    else if( o->t > f->t )
        f->setRight(remove(f->getRight(p),o,p),p);
    else if (o == f) // 
    {
        Fiber* q = f->getLeft(p);
        Fiber* r = f->getRight(p);
        //delete f;
        if( !r ) return q;
        Fiber* min = findmin(r,p);

        min->setRight( removemin(r,p) ,p);
        min->setLeft(q,p);
        return balance(min,p);
    }else{
    	f->setLeft( remove(f->getLeft(p),o,p) ,p);
    	f->setRight(remove(f->getRight(p),o,p),p);
    }
    return balance(f,p);
}

////////////////////////////////////////////////////////////////////////////////////////////////
//              The AVL tree of the perimeters
////////////////////////////////////////////////////////////////////////////////////////////////

Fiber* Perimeter::weakest(){ //
    return findmax(root,this);
}

double Perimeter::weakestLoad(){
    if (root==NULL) return 0;
    return (loadshare()+1)*(findmax(root,this)->t);
}

unsigned char getHeight(Perimeter* p)
{
    return p?p->height:0;
}

int bfactor(Perimeter* p)
{
    return getHeight(p->right)-getHeight(p->left);
}

void fixheight(Perimeter* p)
{
    unsigned char hl = getHeight(p->left);
    unsigned char hr = getHeight(p->right);
    p->height = (hl>hr?hl:hr)+1;
}

Perimeter* rotateright(Perimeter* p)
{
    Perimeter* q = p->left;
    p->left = q->right;
    q->right = p;
    fixheight(p);
    fixheight(q);
    return q;
}

Perimeter* rotateleft(Perimeter* q)
{
    Perimeter* p = q->right;
    q->right = p->left;
    p->left = q;
    fixheight(q);
    fixheight(p);
    return p;
}

Perimeter* balance(Perimeter* p) // balancing the p Perimeter
{
    fixheight(p);
    if( bfactor(p)==2 )
    {
        if( bfactor(p->right) < 0 )
            p->right = rotateright(p->right);
        return rotateleft(p);
    }
    if( bfactor(p)==-2 )
    {
        if( bfactor(p->left) > 0  )
            p->left = rotateleft(p->left);
        return rotateright(p);
    }
    return p; // balancing is not required
}

Perimeter* insert(Perimeter* p, Perimeter* n) // insert k weakestLoad() in a tree with p root
{
    if( !p ) return n;
    if( n->weakestLoad()<p->weakestLoad() )
        p->left = insert(p->left,n);
    else
        p->right = insert(p->right,n);
    return balance(p);
}

Perimeter* findmin(Perimeter* p) // find a Perimeter with minimal key in a p tree 
{
    return p->left?findmin(p->left):p;
}
Perimeter* findmax(Perimeter* p) // find a Perimeter with minimal key in a p tree 
{
    return p->right?findmax(p->right):p;
}
Perimeter* removemin(Perimeter* p) // deleting a Perimeter with minimal key from a p tree
{
    if( p->left==0 )
        return p->right;
    p->left = removemin(p->left);
    return balance(p);
}

Perimeter* remove(Perimeter* p, Perimeter *o) // deleting old Perimeter o from p tree
{

    if( !p ) return 0;
    double oLoad=o->weakestLoad();
    double pLoad=p->weakestLoad();


    if( oLoad < pLoad){
        p->left = remove(p->left,o);
        
    }
    else if( o->weakestLoad() > p->weakestLoad() ){
        p->right = remove(p->right,o);  
    	
    }else if (p==o)
    {
        Perimeter* q = p->left;
        Perimeter* r = p->right;
        if( !r ) return q;
        Perimeter* min = findmin(r);
        min->right = removemin(r);
        min->left = q;
        return balance(min);
    } else
    {
    	p->left = remove(p->left,o);
    	p->right = remove(p->right,o);
    }
    return balance(p);
}

int countThePerimeters(Perimeter *p)
{
    return ( p->left ? countThePerimeters(p->left) : 0 ) + ( p->right ? countThePerimeters(p->right) : 0 ) + 1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//              The AVL tree for fibers in multiple perimeters
//////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char getHeight(ManyPerimeter* mp)
{
    return mp?mp->height:0;
}

int bfactor(ManyPerimeter* mp)
{
    return getHeight(mp->right)-getHeight(mp->left);
}

void fixheight(ManyPerimeter* mp)
{
    unsigned char hl = getHeight(mp->left);
    unsigned char hr = getHeight(mp->right);
    mp->height = (hl>hr?hl:hr)+1;
}

ManyPerimeter* rotateright(ManyPerimeter* mp)
{
    ManyPerimeter* mq = mp->left;
    mp->left = mq->right;
    mq->right = mp;
    fixheight(mp);
    fixheight(mq);
    return mq;
}

ManyPerimeter* rotateleft(ManyPerimeter* mq)
{
    ManyPerimeter* mp = mq->right;
    mq->right = mp->left;
    mp->left = mq;
    fixheight(mq);
    fixheight(mp);
    return mp;
}

ManyPerimeter* balance(ManyPerimeter* mp) // balancing the p Perimeter
{
    fixheight(mp);
    if( bfactor(mp)==2 )
    {
        if( bfactor(mp->right) < 0 )
            mp->right = rotateright(mp->right);
        return rotateleft(mp);
    }
    if( bfactor(mp)==-2 )
    {
        if( bfactor(mp->left) > 0  )
            mp->left = rotateleft(mp->left);
        return rotateright(mp);
    }
    return mp; // balancing is not required
}

ManyPerimeter* insert(ManyPerimeter* mp, ManyPerimeter* mn) // insert k weakestLoad() in a tree with p root
{
    if( !mp ) return mn;
    if( mn->trueload < mp->trueload )
        mp->left = insert(mp->left,mn);
    else
        mp->right = insert(mp->right,mn);
    return balance(mp);
}

ManyPerimeter* findmin(ManyPerimeter* mp) // find a Perimeter with minimal key in a p tree 
{
    return mp->left?findmin(mp->left):mp;
}

ManyPerimeter* findmax(ManyPerimeter* mp) // find a Perimeter with minimal key in a p tree 
{
    return mp->right?findmax(mp->right):mp;
}
ManyPerimeter* removemin(ManyPerimeter* mp) // deleting a Perimeter with minimal key from a p tree
{
    if( mp->left==0 )
        return mp->right;
    mp->left = removemin(mp->left);
    return balance(mp);
}

ManyPerimeter* remove(ManyPerimeter* mp, ManyPerimeter *mo) // deleting old Perimeter o from p tree
{
    if( !mp ) return 0;
    if(!mo ){printf("tried to remove NULL pointer! \n"); exit(-1);}

    if( mo->trueload < mp->trueload){
        mp->left = remove(mp->left,mo);
        
    }
    else if( mo->trueload > mp->trueload ){
        mp->right = remove(mp->right,mo);  
        
    }else if (mp==mo)
    {
        ManyPerimeter* mq = mp->left;
        ManyPerimeter* mr = mp->right;
        if( !mr ) return mq;
        ManyPerimeter* min = findmin(mr);
        min->right = removemin(mr);
        min->left = mq;
        return balance(min);
    } else
    {
        mp->left = remove(mp->left,mo);
        mp->right = remove(mp->right,mo);
    }
    return balance(mp);
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//              The linked list of many perimeters, linkes all fibers within a perimeter included in many perimeters
/////////////////////////////////////////////////////////////////////////////////////////////////
void insert(Fiber *f, Perimeter *p){    //insert fiber f in manyperimeterlist associated with perimeter p
    ManyPerimeterLinkedList *n = new ManyPerimeterLinkedList;
    n->f=f;
    ManyPerimeterLinkedList *h=p->manyPerHead;
    if(!h){//New head
        p->manyPerHead=n;
        return;
    }else{  //found head, insert behind head
        n->next=h;
        p->manyPerHead=n;
    }
}

void remove(ManyPerimeterLinkedList *o, ManyPerimeterLinkedList *prv, Perimeter *p){ //remove o from the manyperimeterlinkedlist associated with perimeter p
    if(!o){
        printf("NULL in remove\n");
        exit(-1);
    }
    ManyPerimeterLinkedList *nxt=o->next;
    if(!prv){//new head
        p->manyPerHead=nxt;
    }else{
        prv->next=nxt;
    }
    delete o;
}

void deleteList(Perimeter *p){
    ManyPerimeterLinkedList *m=p->manyPerHead;
    ManyPerimeterLinkedList *n;
    while (m){   
        n=m->next;
        delete m;
        m=n;
    }
    p->manyPerHead=NULL;
}