#include <GLFW/glfw3.h> 
#include "stb_easy_font.h"  //Found in the same folder as the current file.

int thePerimeterCount=0;    
double PI = M_PI;
GLFWwindow* window;
double prevx, prevy, x, y, resx, resy, height, width;
int clickedButtons = 0;
bool changed = false;
enum buttonMaps { FIRST_BUTTON=1, SECOND_BUTTON=2, THIRD_BUTTON=4, FOURTH_BUTTON=8, FIFTH_BUTTON=16, NO_BUTTON=0 };
enum modifierMaps { CTRL=2, SHIFT=1, ALT=4, META=8, NO_MODIFIER=0 };
void runSimulation();
void initGL();
void Draw();
void t_callback(GLFWwindow* win, int t, int scancode, int action, int mods);
void mousebutton_callback(GLFWwindow* win, int button, int action, int mods);
void mousepos_callback(GLFWwindow* win, double xpos, double ypos);
void mousewheel_callback(GLFWwindow* win, double xoffset, double yoffset);
void print_string(float t, float y, char *text, float r, float g, float b);
enum drawEnum {DRAW_STATUS,DRAW_PERIMETER,DRAW_EFFECTIVE_THRESHOLD,DRAW_FORCE,DRAW_NUM};
int DRAW_STYLE = DRAW_STATUS;
void breakWeakest();
void break_ELS();
//Print a string with screen coordinates (Fixed size and position)
void print_string(float t, float y, char *text, float r, float g, float b)
{
  static char buffer[99999]; // ~500 chars
  int num_quads;
  num_quads = stb_easy_font_print(t, y, text, NULL, buffer, sizeof(buffer));
  glColor3f(r,g,b);
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(2, GL_FLOAT, 16, buffer);
  glDrawArrays(GL_QUADS, 0, num_quads*4);
  glDisableClientState(GL_VERTEX_ARRAY);
}
//Print a string to world coordinates (On canvas)
void print_string_world(float xx, float yy, char *text, float r, float g, float b, float scale) {
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,resx,resy,0, -1000, 1000);
 
    glTranslatef((xx - x)/width * resx, (y + height - yy)/height * resy , 0);
    glScalef(scale, scale, scale);
    print_string(0,0,text,r, g, b);
    glPopMatrix();
}
void runWithGraphics(){
    prevx = -1, prevy = -1;
    x = -1, y = -1;
    resx=1000, resy=700;
    height=Lx+3, width=height*resx/(double)resy;
    initGL();
    runSimulation();
    while ( !glfwWindowShouldClose(window)) {
      int display_w;
      int display_h;
      glLoadIdentity();
      glfwGetFramebufferSize(window, &display_w, &display_h);
      resx=display_w/2;
      resy=display_h/2;
      width=height*resx/(double)resy;
      glViewport(0, 0, display_w, display_h);
      glOrtho(x, x + width, y, y + height, -1000, 1000);
      Draw();
      glfwPollEvents();
    }
}
void runSimulation(){
  printf("System size %d t %d (%d) \n", Lx,Ly, N);
    printf("Cutoff %f \n", cutoff);
  fibers=new Fiber[N];
  int xx;
  int yy;
  //allocate noperimeter
  noperimeter.resize(N);
  double threshold;
  for (int i = 0; i < N; i++) {
    xx = i % Lx;
    yy = i / Lx;
    threshold= cutoff - log(rng());
    fibers[i].t = 1.0/threshold; //store the inverse fiber threshold
    fibers[i].fiberID=i;
        noperimeter[i]=&fibers[i];
    for (int j = 0; j < 4; ++j)
    {
      fibers[i].height[j]=1;
      fibers[i].left[j]=NULL;
      fibers[i].right[j]=NULL;
      fibers[i].myPerimeter[j]=NULL;
    }

  }
  std::sort(noperimeter.begin(), noperimeter.end(), lowestThreshold); //Sort fibers in noperimeter wrt. the true theshold 
  printf("Strongest sigma/x %f, weakest sigma/x is %f \n",noperimeter.front()->t,noperimeter.back()->t);
}
void drawFilledCircle(GLfloat x, GLfloat y, GLfloat radius,GLfloat ratio){
    GLfloat center = 128.0f;
    GLfloat width = 127.0f;
    GLfloat twicePi = 2.0f * PI;
    GLfloat red,green,blue;
    red   = sin(twicePi*ratio+2); 
    green = sin(twicePi*ratio+0); 
    blue  = sin(twicePi*ratio+4); 
    int i;
    int triangleAmount = 5; //# of triangles used to draw circle    
    glColor3f(red,green,blue);
    glBegin(GL_TRIANGLE_FAN);
        glVertex2f(x, y); // center of circle
        for(i = 0; i <= triangleAmount;i++) { 
            glVertex2f(
                    x + (radius * cos(i *  twicePi / triangleAmount)), 
                y + (radius * sin(i * twicePi / triangleAmount))
            );
        }
    glEnd();
}
void preOrderDraw(Fiber *f, Perimeter* p,double ratio){
  
  GLfloat twicePi = 2.0f * PI;
  GLfloat red,green,blue;
   
  red   = sin(twicePi*ratio+2); 
  green = sin(twicePi*ratio+0); 
  blue  = sin(twicePi*ratio+4); 

  if(f==NULL) return;
  GLfloat fx=floor(f->fiberID % Lx)+0.5+0.2*ratio;
  GLfloat fy=floor(f->fiberID / Lx)+0.5+0.2*ratio;

  GLfloat r=0.5;
  int n=10;
  glColor3f(red,green,blue);
  glBegin(GL_TRIANGLE_STRIP);
  for (int i = 0; i <= n+1; i++)
      glVertex2f( fx+r*cos(2*PI*i/n), fy+r*sin(2*PI*i/n) );
  glEnd();

  preOrderDraw(f->getLeft(p),p,ratio);
  preOrderDraw(f->getRight(p),p,ratio);

}
void drawNodes(Fiber *f, Perimeter *p, double x, double y, int level, GLfloat ratio) {
    if (f != NULL) { 
      GLfloat twicePi = 2.0f * PI;
      GLfloat red,green,blue;
     
      red   = sin(twicePi*ratio+2); 
      green = sin(twicePi*ratio+0); 
      blue  = sin(twicePi*ratio+4); 

      float dx = 1.0/(1 << level);
      float dy = 2.0*dx;
 
      drawNodes(f->getLeft(p),p,  x - dx, y - dy, level+1,ratio);
      drawNodes(f->getRight(p),p, x + dx, y - dy, level+1,ratio);
 
      char str[80];
      sprintf(str, "%.3f", f->t);
      int text_width = stb_easy_font_width(str);
      int text_height = 8.0;
      
      // "shifting to get text centered, scale worked by triangle and error. probably a good reason for these numbers"
      print_string_world(x - 0.025*dx*text_width/2, y + 0.025*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
      sprintf(str, "fiber %d, height %d ", f->fiberID,getHeight(f,p));
      print_string_world(x - 0.035*dx*text_width/2, y +0.3*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
      glColor3f(red,green,blue);
      if (f->getLeft(p)) {
          glBegin(GL_LINES);
          glVertex2f(x,y);
          glVertex2f(x - dx,y-dy);
          glEnd(); 
      }
      if (f->getRight(p)) {
          glBegin(GL_LINES);
          glVertex2f(x,y);
          glVertex2f(x + dx,y-dy);
          glEnd();
      }
      
      glBegin(GL_TRIANGLE_STRIP);
      double r = dx;
      for (int i = 0; i <= 11; i++) {
          glVertex2f(x+r*cos(2*PI*i/10.0), y+r*sin(2*PI*i/10.0));
      }
      glEnd();
    }
}
void drawTrees(Perimeter *p, double x, double y, int level, int &count) {
    if (p != NULL) {
      GLfloat ratio =0;
      GLfloat twicePi = 2.0f * PI;
      GLfloat red,green,blue=0;
      float dx = 1.0/(1 << level);
      float dy = 2.0*dx;
      char str[80];
      sprintf(str, "%.3f", p->weakestLoad());
      int text_width = stb_easy_font_width(str);
      int text_height = 8.0;
      // "shifting to get text centered, scale worked by triangle and error. Probably a good reason for these numbers (sic)"
      print_string_world(x - 0.025*dx*text_width/2, y + 0.025*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
      sprintf(str, "fiber %d, height %d ", p->weakest()->fiberID,getHeight(p));
      print_string_world(x - 0.035*dx*text_width/2, y +0.3*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
      if (p->left) {
          glBegin(GL_LINES);
          glVertex2f(x,y);
          glVertex2f(x - dx,y-dy);
          glEnd();
      }
      if (p->right) {

          glBegin(GL_LINES);
          glVertex2f(x,y);
          glVertex2f(x + dx,y-dy);
          glEnd();
      }     
      ratio=count/(float)thePerimeterCount;
      count+=1;
      red   = sin(twicePi*ratio+2); 
      green = sin(twicePi*ratio+0); 
      blue  = sin(twicePi*ratio+4); 
      glColor3f(red,green,blue);

      glBegin(GL_TRIANGLE_STRIP);

      double r = dx;
      for (int i = 0; i <= 11; i++) {
          glVertex2f(x+r*cos(2*PI*i/10.0), y+r*sin(2*PI*i/10.0));
      }
      glEnd();
      drawTrees(p->left,  x - dx, y - dy, level+1,count);//,ratio);
      drawTrees(p->right, x + dx, y - dy, level+1,count);//,ratio);
    }
 
}
void drawManyPerimeterTree(ManyPerimeter *mp, double x, double y, int level, int &count) {
    if (mp != NULL) {
        GLfloat ratio =0;
        GLfloat twicePi = 2.0f * PI;
        GLfloat red,green,blue=0;
        float dx = 1.0/(1 << level);
        float dy = 2.0*dx;
        char str[80];
        sprintf(str, "%.3f", mp->trueload);
        int text_width = stb_easy_font_width(str);
        int text_height = 8.0;
        // shifting to get text centered, scale worked by triangle and error. probably a good reason for these numbers
        print_string_world(x - 0.025*dx*text_width/2, y + 0.025*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
        sprintf(str, "fiber %d, height %d ", mp->f->fiberID,getHeight(mp));
        print_string_world(x - 0.035*dx*text_width/2, y +0.3*dx*text_height/2, str, 0.0, 0.0, 0.0, 0.025*dx*resx/width);
        //glColor3f(red,green,blue);
        if (mp->left) {
            glBegin(GL_LINES);
            glVertex2f(x,y);
            glVertex2f(x - dx,y-dy);
            glEnd(); 
        }
        if (mp->right) {
            glBegin(GL_LINES);
            glVertex2f(x,y);
            glVertex2f(x + dx,y-dy);
            glEnd();
        }
        count+=1;
        glColor3f(red,green,blue);
        glBegin(GL_TRIANGLE_STRIP);
        double r = dx;
        for (int i = 0; i <= 11; i++) {
            glVertex2f(x+r*cos(2*PI*i/10.0), y+r*sin(2*PI*i/10.0));
        }
        glEnd();
        drawManyPerimeterTree(mp->left,  x - dx, y - dy, level+1,count);//,ratio);
        drawManyPerimeterTree(mp->right, x + dx, y - dy, level+1,count);//,ratio);
    }
}
void drawPerimeters(Perimeter* p,int &count){
    if (p != NULL){

    float ratio=count/(float)thePerimeterCount;
    count++;
    preOrderDraw(p->root,p,ratio);
      drawPerimeters(p->left,count);
      drawPerimeters(p->right,count);
  }
}
void drawTree(Perimeter *p,int &m){
  if(p!=NULL){
    float ratio=m/(float)thePerimeterCount;
    drawNodes(p->root,p,1+3.5*m,7,0,ratio);
    m++;
    drawTree(p->left,m);
    drawTree(p->right,m);
  }
}
void drawManyPers(ManyPerimeter *mp){
    if(mp==NULL) return;
    GLfloat twicePi = 2.0f * PI;
    GLfloat red,green,blue;
    GLfloat ratio=3.14;
     
    red   = sin(twicePi*ratio+2); 
    green = sin(twicePi*ratio+0); 
    blue  = sin(twicePi*ratio+4); 

    
    GLfloat fx=floor(mp->f->fiberID % Lx)+0.5;//+0.2*ratio;
    GLfloat fy=floor(mp->f->fiberID / Lx)+0.5;//+0.2*ratio;
    
    drawFilledCircle(fx, fy, 0.25,ratio);//);//,ratio);
    drawManyPers(mp->left);//,ratio);
    drawManyPers(mp->right);//,ratio);

}
void drawManyPersVector(){
    std::vector<Fiber *>::iterator it;
    it=manyPerimeters.begin();
    while(it!=manyPerimeters.end()){

      if((*it)->status==BROKEN || (*it)->perimeterCount() < 2){
          manyPerimeters.erase(it);
          //printf("done?\n");
      }else{
            GLfloat twicePi = 2.0f * PI;
            GLfloat red,green,blue;
            GLfloat ratio=3.14;
             
            red   = sin(twicePi*ratio+2); 
            green = sin(twicePi*ratio+0); 
            blue  = sin(twicePi*ratio+4); 
        
            
            GLfloat fx=floor((*it)->fiberID % Lx)+0.5;//+0.2*ratio;
            GLfloat fy=floor((*it)->fiberID / Lx)+0.5;//+0.2*ratio;
            drawFilledCircle(fx, fy, 0.25,ratio);//);//,ratio);
            ++it;
        }
    }
}
int totPerCount(Perimeter *p){
    return ( p->left ? totPerCount(p->left) : 0 ) + ( p->right ? totPerCount(p->right) : 0 ) + p->intactFibers;
}
int btotPerCount(Perimeter *p){
    return ( p->left ? btotPerCount(p->left) : 0 ) + ( p->right ? btotPerCount(p->right) : 0 ) + p->brokenFibers;
}
int countManyPerimeterNodes(ManyPerimeter *mp)
{
    return ( mp->left ? countManyPerimeterNodes(mp->left) : 0 ) + ( mp->right ? countManyPerimeterNodes(mp->right) : 0 ) + 1;
}
void Draw() {
  static int frame = 0;
  static char str[256];
  static double t2 = glfwGetTime();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  Perimeter *myPer;
  Fiber *f;
  Perimeter *p;
  GLfloat fx;
  GLfloat fy;
  GLfloat px;
  GLfloat py;
  p=rootPerimeter;
  int fibersInPer=0;
  if(rootPerimeter) thePerimeterCount=countThePerimeters(p);

  int count=0;
  GLfloat ratio=0;
  
  if(DRAW_STYLE==DRAW_STATUS and rootPerimeter!=NULL){

    if(totBreaks != N) drawPerimeters(p,count);
        
        if(totBreaks<(N*treeToVec)){
            if(manyPerimeterRoot) drawManyPers(manyPerimeterRoot);
        }else{
            drawManyPersVector();
        }
  }
  count=0;

  if(DRAW_STYLE==DRAW_PERIMETER and rootPerimeter!=NULL) {
    drawTree(p,count);
    count=0;
    drawTrees(p,5,2,0,count);
  }
    
  else{
    for (int i = 0; i < Lx; i++) { 
      for (int j = 0; j < Ly; j++) {
        int k = j*Lx + i;
  
        if (DRAW_STYLE == DRAW_STATUS) {
  
          if (fibers[k].status == BROKEN) {
            glColor3f(0,0,0);
          } else{
            glColor3f(0.95,0.95,0.95);
          }
        } 
        glBegin(GL_QUADS);
        double len = 0.98;
        glVertex2f(i,       j);
        glVertex2f(i + len, j);
        glVertex2f(i + len, j + len);
        glVertex2f(i,       j + len);
        glEnd();
      }
    }   
  }
    // Text
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,resx,resy,0, -1000, 1000);
    glScalef(1.50,1.50,1.50);
    glTranslatef(3,3,0);
    glColor3f(0.5,0.5,0.5);
    int j=0;   
    GLfloat red,green,blue;
    GLfloat twicePi=2.0f*PI;
    double multi_loadshare=0;
    double multiPerLoad;

    if(DRAW_STYLE==DRAW_PERIMETER and rootPerimeter!= NULL){
      sprintf(str,"The root perimeter has root with fiber ID %d, and has %d intact fibers", rootPerimeter->root->fiberID,rootPerimeter->intactFibers);
      print_string(3,20+10,str,1,0,0);
    }
    
    if(totBreaks==N){
      sprintf(str,"ALL FIBERS ARE BROKEN!!!");
      print_string(3,20+10*(j++ + 1),str,1,0,0);
    }
    int perc=rootPerimeter?totPerCount(rootPerimeter):0;
    int bperc=rootPerimeter?btotPerCount(rootPerimeter):0;
    int mperc;
    if(totBreaks<(N*treeToVec)){ 
        mperc=manyPerimeterRoot?countManyPerimeterNodes(manyPerimeterRoot):0;
    }else{
            mperc=manyPerimeters.size();
        }
        double t3 = glfwGetTime();
        sprintf(str,"Frame #%d, Fiber: %d, Broken fibers: %d (%.2f %%), Perimeters:  %d, Fibers in perimeters: %d, Fibers in manyPerimeters: %d,  time_since_start = %.1fs (%.1fs)", 
            frame++,N,totBreaks,100*(double) totBreaks/(double) N,thePerimeterCount,perc,mperc,glfwGetTime()-time_at_start,t3-t2); 
    print_string(0,0,str,0,0,0);
    glPopMatrix();
  glfwSwapBuffers(window);
}

void initGL() {
  printf("Initializing OpenGL/GLFW"); fflush(stdout);
  if (!glfwInit()) {
    printf("Could not initialize\n");
    exit(-1);
  }

  glfwWindowHint(GLFW_SAMPLES , 2);
  window = glfwCreateWindow(resx, resy, "test", 0, 0);
  if (!window) {
    printf("Could not open glfw window\n");
    glfwTerminate();
    exit(-2);
  }
  glfwMakeContextCurrent(window);

  glEnable(GL_DEPTH_TEST);
  glOrtho(x, x + width, y, y + height, -1000, 1000);
  glClearColor(1,1,1,1);
  glfwSwapInterval(1);

  glfwSetKeyCallback(window, t_callback);
  glfwSetMouseButtonCallback(window, mousebutton_callback);
  glfwSetScrollCallback(window, mousewheel_callback);
  glfwSetCursorPosCallback(window, mousepos_callback);

  printf(" ... DONE\n");
}


void t_callback(GLFWwindow* win, int t, int scancode, int action, int mods) {

  if (t == GLFW_KEY_ESCAPE) {
    glfwSetWindowShouldClose(win, GL_TRUE);
  }
  if(t == GLFW_KEY_SPACE && action ){
    for(int i=0;i<(N/100);i++){
      if(totBreaks==N) return;
      breakWeakest();
    }
  }
  if(t == GLFW_KEY_R && action ){
    for(int i=0;i<(N);i++){
      if(totBreaks % 200==0) Draw();
      if(totBreaks==N) return;
      breakWeakest();
    }
  }
  if(t == GLFW_KEY_O && action ){
    if(totBreaks==N) return;
    breakWeakest();
  }
  if(t == GLFW_KEY_E && action ){
    if(totBreaks==N) return;
    break_ELS();
  }
  if(t == GLFW_KEY_TAB && action ){
    for(int i=0;i<(N/10);i++){
      if(totBreaks==N) return;
      break_ELS();
    }
  }
  if (t == '1' && action) {
    DRAW_STYLE = DRAW_STATUS;
  }
  if (t == '2' && action) {
    DRAW_STYLE = DRAW_PERIMETER;
  }
}
void mousebutton_callback(GLFWwindow* win, int button, int action, int mods) {
  glfwGetCursorPos(win,&prevx,&prevy);
  double xend = x + prevx*(width)/resx;
  double yend = y + (resy-prevy)*height/resy;
  if (action == 1)
    clickedButtons |= (1 << button);
  else
    clickedButtons &= ~(1 << button);
  if (button == GLFW_MOUSE_BUTTON_1 && action) {
  } else if (button == GLFW_MOUSE_BUTTON_2 && action) {
    if (xend >= 0 && xend < Lx && yend >= 0 && yend < Ly) {
      int k = int(yend)*Lx + int(xend);
      fibers[k].Break();
    }
  }
  fflush(stdout);
}
void mousepos_callback(GLFWwindow* win, double xpos, double ypos) {
  double xend = x + xpos*(width)/resx;
  double yend = y + (resy-ypos)*height/resy;

  if (clickedButtons&FIRST_BUTTON) {
    // Moving camera
    x -= (xpos-prevx)/resx*width;
    y += (ypos-prevy)/resy*height;

    prevx = xpos;
    prevy = ypos;
    changed = true;
  } else if (clickedButtons&SECOND_BUTTON) {
    // "Drawing"
    if (xend >= 0 && xend < Lx && yend >= 0 && yend < Ly) {
      int k = int(yend)*Lx + int(xend);
      fibers[k].Break();
            if(totBreaks>=int(N*treeToVec)){
                updateManyPerVector();
            }
    }
  }
}
void mousewheel_callback(GLFWwindow* win, double xoffset, double yoffset)
{
  double zoomFactor = pow(0.95,yoffset);
  x = x + width*(1.0 - zoomFactor)/2.0;
  y = y + height*(1.0 - zoomFactor)/2.0;
  width *= zoomFactor;
  height *= zoomFactor;
  changed = true;
}